#pylint: disable=consider-using-f-string
from argparse import Namespace as ArgparseNamespace
from pathlib import Path
from tempfile import TemporaryDirectory
import os

from beets import config as beetsconfig
from beets.ui import _setup as setup_beets
from beets.ui.commands import import_files

from ytbdl import beetsplug
from ytbdl import config, config_exists, get_main_config_path
from ytbdl.exceptions import ConfigurationError


# Keys and their required content in the config file
REQUIRED_VARIABLES = (
    ('directory', r'{import_dir}'),
    ('pluginpath', r'{beetsplug_dir}'),
)

# Plugins required in plugins: list in the config file
REQUIRED_PLUGINS = (
    'fromdirname',
    'fromyoutubetitle',
)


def beet_import(album_dir: Path, logger):
    ''' Emulates the behaviour of calling Beets' import function from a shell,
    in an embedded fashion. This bypasses a lot of the overhead required in
    creating a new subprocess as well as for other set up. Since the default
    beets config and library are used, no custom processing is required to set
    those up.

    Behind the scenes, beets will open a
    :code:`beets.ui.commands.TerminalImportSession` so that users can enter
    input via stdin. That function will emit the cli_exit event in case the user
    has activated any plugins that rely on this event. All of this functionality
    is emulated here.

    Args:
        album_dir (Path): A path to an album directory where some music exists
        logger: A logging object
    '''
    import_dir = str(Path(album_dir).parent.parent.resolve()).replace('\\', '/')
    beetsplug_dir = str(Path(beetsplug.__file__).parent.resolve()).replace('\\', '/')

    config_content = get_custom_config_contents(
        import_dir=import_dir,
        beetsplug_dir=beetsplug_dir
    )

    with TemporaryDirectory() as temp_dir:
        temp_file_path = os.path.join(temp_dir, 'config.yaml')

        with open(temp_file_path, 'w', encoding='utf-8') as temp_config:
            temp_config.write(config_content)

        logger.debug(msg=('Created a temporary beets config at: {0}'.format(
            temp_file_path
        )))

        setup_options = ArgparseNamespace(
            directory=None,
            config=temp_file_path,
            plugins=None,
            library=None,
        )

        _, plugins, library = setup_beets(setup_options)

        # Start the import
        paths = [str(album_dir).encode('utf-8')]
        import_files(library, paths, None)

        # Clean up
        plugins.send('cli_exit', lib=library)
        library._close()
        beetsconfig.clear()


def get_custom_config_contents(**template_args) -> str:
    ''' Get the content of the custom beets configuration specified in ytbdl's
    config, and verify that the options with "DO NOT REMOVE" were not removed by
    the user.

    Args:
        **template_args: Supplied arguments to fill ytbdl's string template

    Returns:
        (str): ytbdl's config, with filled template arguments
    '''
    if not config_exists():
        raise ConfigurationError('Could not find a config file')

    for required_key, required_content in REQUIRED_VARIABLES:
        if required_key not in config:
            raise ConfigurationError(
                'The "{0}" key is missing from the configuration file. Add it '
                'back, and set it to: "{1}" (keep the quotes)'.format(
                    required_key, required_content
                )
            )
        found_content = config[required_key].as_str()
        if found_content != required_content:
            raise ConfigurationError(
                'The "{0}" key in the configuration file must be set to '
                '"{1}," but found "{2}"'.format(
                    required_key, required_content, found_content
                )
            )

    plugin_list = config['plugins'].get(list)
    for required_plugin in REQUIRED_PLUGINS:
        if required_plugin not in plugin_list:
            raise ConfigurationError(
                'The "{0}" plugin is missing from the plugins list in the '
                'configuration file. Add it back before continuing'.format(
                    required_plugin
                )
            )

    raw_beets_template = ''
    with open(get_main_config_path(), 'r', encoding='utf-8') as file_pointer:
        raw_beets_template = file_pointer.read()

    return raw_beets_template.format(**template_args)



#pylint: disable=consider-using-f-string
from pathlib import Path
from subprocess import CalledProcessError
import re
import sys

import confuse

from ytbdl import config_exists, config
from ytbdl.apps.base import BaseApp
from ytbdl.beets import beet_import
from ytbdl.exceptions import ConfigurationError
from ytbdl.yt_dlp import ytdl_options, download_audio


class DownloadApp(BaseApp):
    ''' App to download and tag audio files downloaded from the internet.
    '''

    @staticmethod
    def add_sub_parser_arguments(sub_parser):
        dl_parser = sub_parser.add_parser(name='get', description=(
            'download one or more audio files with yt-dlp before tagging them '
            'as an album with beets. beets\' behaviour can be changed by '
            'modifying the ytbdl config file, which is itself a beets config '
            'file. the config file also has a ytdl_args option that can be '
            'used to control yt-dlp\'s behaviour (or use --ytdl-args)'
        ))
        dl_parser.add_argument('-v', '--verbose', action='store_true', help=(
            'log verbose (debug) information'
        ))
        dl_parser.add_argument('-y', '--ytdl-args', default=[],
            type=ytdl_options, help=(
            'command line arguments to pass to yt-dlp. For example: '
            '"--format bestaudio[ext=m4a]". you may also specify yt-dlp '
            'options in the ytdl_args setting in the config file. you may '
            'not use -x/--extract-audio or -o/--output as these are already in '
            'use. --ytdl-args are always combined with any existing args in '
            'the ytdl_args config option'
        ))
        dl_parser.add_argument('artist', help=(
            'the artist who created the album'
        ))
        dl_parser.add_argument('album', help=(
            'the name of the album to download'
        ))
        dl_parser.add_argument('urls', nargs='+', help=(
            'one or more URLs to download audio from'
        ))

    INVALID_FILENAME_CHARS = re.compile(r'[^\w\-_\. ]')

    def __init__(self):
        self.verbose = False
        self.logger = None
        self.temp_config = None

    def configure_logging(self):
        level = 'DEBUG' if self.verbose else 'INFO'
        self.logger = self.get_logger('ytbdl', level)

    def start_execution(self, arg_parser, **kwargs):
        self.verbose = kwargs.get('verbose')
        self.configure_logging()
        if not config_exists():
            self.logger.info('Create a config before continuing with:')
            self.logger.info('ytbdl config create')
            return

        artist_name = kwargs.get('artist')
        album_name = kwargs.get('album')
        urls = kwargs.get('urls')
        extra_args = kwargs.get('ytdl_args', [])

        try:
            # Construct yt-dlp extra arguments
            if 'ytdl_args' in config:
                for config_arg in config['ytdl_args'].get(list):
                    if config_arg not in extra_args:
                        extra_args.append(config_arg)
            else:
                self.logger.debug('ytdl_args not found in config file')

            album_dir = self.get_album_dir(artist_name, album_name)

            # Download music to directory (yt-dlp will create the directory if
            # it's missing)
            self.logger.info(msg='Downloading "{0}" by {1}'.format(
                album_name, artist_name
            ))
            download_audio(album_dir, extra_args, urls, self.logger)

            # Autotag music in directory
            self.logger.info(msg='Autotagging album downloaded to {0}'.format(
                str(album_dir)
            ))
            beet_import(album_dir, self.logger)

        except confuse.exceptions.ConfigTypeError:
            self.logger.error('ytdl_args config option is not a list!')
            self.logger.warning('Aborting')
            sys.exit(1)
        except KeyboardInterrupt:
            self.logger.info('User interrupted program.')
            self.logger.info('Aborting.')
            sys.exit(2)
        except FileExistsError as exc:
            self.logger.error(msg='FileExistsError: {0}'.format(str(exc)))
            self.logger.warning('Aborting')
            sys.exit(1)
        except (CalledProcessError, ConfigurationError) as exc:
            self.logger.error(msg='{0} encountered:'.format(
                exc.__class__.__name__
            ))
            self.logger.error(msg=str(exc))
            self.logger.warning('Aborting')
            sys.exit(1)


    def get_album_dir(self, artist: str, album: str) -> Path:
        ''' Get the path to the artist/album folder. If the album folder already
        exists and is not empty, an exception is raised as this may indicate
        that the album has already been downloaded.

        The directory structure is created under the current directory in this
        manner:

        .. code-block::

            . (current directory)
                |_ Artist
                    |_ Album

        Args:
            artist (str): The name of the artist whose music is being downloaded
            album (str): The name of album by the artist

        Returns:
            (Path): A path to the album folder relative to the current directory
        '''
        def sanitize_path(name):
            return self.INVALID_FILENAME_CHARS.sub('_', name)

        album_folder = Path('.') / sanitize_path(artist) / sanitize_path(album)
        if album_folder.exists() and album_folder.is_dir() and \
            any(album_folder.glob('*')):
            raise FileExistsError(
                'The album folder "{0}" already exists and is not empty'.format(
                    str(album_folder)
                )
            )
        return album_folder



import argparse

from .apps.config import ConfigApp
from .apps.get import DownloadApp

ACTIVATED_APPS = {
    'config': ConfigApp,
    'get': DownloadApp,
}

def main():
    arg_parser = get_app_arg_parser()
    parsed_namespace = arg_parser.parse_args()
    arguments = vars(parsed_namespace)
    application = ACTIVATED_APPS[arguments['sub-app']]()
    application.start_execution(arg_parser, **arguments)

def get_app_arg_parser():
    app_parser = argparse.ArgumentParser(description=(
        'download songs with yt-dlp and auto-tag them with beets. use the get '
        'sub-app to *get* music, and use the config sub-app to *config*ure '
        'yt-dlp\'s and beets\' behaviour'
    ))
    subparser = app_parser.add_subparsers(title='Sub-application Choice')
    subparser.required = True
    subparser.dest = 'sub-app'
    for _, class_ in ACTIVATED_APPS.items():
        class_.add_sub_parser_arguments(subparser)
    return app_parser

if __name__ == '__main__':
    main()



import os

import confuse

config = confuse.LazyConfig('ytbdl', None)

def get_loaded_config_sources():
    ''' Get existing configuration files

    Returns:
        (list): A list of (string) paths to configuration files that exist on
            the file system. Returns an empty list if no configuration files
            exist
    '''
    config.resolve()
    return [s.filename for s in config.sources if os.path.exists(s.filename)]

def get_main_config_path():
    ''' Get the main configuration file path

    Returns:
        (str): A path to the configuration file. This path may or may not exist
    '''
    return os.path.join(config.config_dir(), 'config.yaml')

def config_exists():
    ''' Determine if one or more configuration files exist.

    Returns:
        (bool): True if a config file exists, False otherwise
    '''
    return any(get_loaded_config_sources())



"""If the album and artist are empty, try to extract the album and artist from
the directories the file is in. Looks above the file for the album, and above
the album for the artist name.
"""

from pathlib import Path

from beets.plugins import BeetsPlugin
from beets.util import displayable_path

import tagsfrompath as frompath


class FromDirectoryNamePlugin(BeetsPlugin):
    def __init__(self):
        super(FromDirectoryNamePlugin, self).__init__()
        self.register_listener('import_task_start', update_album_artist_with_dirnames)


def update_album_artist_with_dirnames(task, session):
    items = task.items if task.is_album else [task.item]

    for item in items:
        if item.album and item.artist:
            continue

        file_path = Path(displayable_path(item.path))

        if not item.album:
            item.album = frompath.get_album_name(file_path)
        if not item.artist:
            item.artist = frompath.get_artist_name(file_path)



""" fromyoutubetitle Beets Plugin """

from pathlib import Path
import re

from beets.plugins import BeetsPlugin
from beets.util import displayable_path

import tagsfrompath as frompath


class FromYoutubeTitlePlugin(BeetsPlugin):
    """ Sets the title of each item to the filename, removing most of the common
    junk associated with YouTube titles like "(Official Audio)" and the name of
    the album or artist.
    Assumes the music is in an Artist/Album/Song folder structure, and that the
    song file names are the names of the YouTube videos they were extracted
    from.
    """
    def __init__(self):
        super(FromYoutubeTitlePlugin, self).__init__()
        self.register_listener('import_task_start', set_titles_no_junk)


YOUTUBE_TITLE_JUNK = [
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?(?:Explicit|Clean|Parental\sAdvisory).*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?(?:HQ|HD|CDQ).*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Audio.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Album.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Song.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Video.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Lyric.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Visualizer.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?iTunes.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Official.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Original.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Version.*?[\)\]\}])'),
    re.compile(r'(?i)(?P<junk>[\(\[\{].*?Prod(?:uced|\.)?\sBy.*?[\)\]\}])'),
]


EXTRA_STRIP_PATTERNS = [
    re.compile(r'^\s*[-_\|]\s*(?P<title>.+)$'),
    re.compile(r'^(?P<title>.+)\s*[-_\|]\s*$')
]


def set_titles_no_junk(task, session):
    items = task.items if task.is_album else [task.item]

    for item in items:
        if item.title:
            continue
        item_file_path = Path(displayable_path(item.path))
        youtube_title = frompath.get_title(item_file_path)
        album_name = frompath.get_album_name(item_file_path)
        artist_name = frompath.get_artist_name(item_file_path)
        artist_album_junk = [
            '(?i)(?P<junk>\\({0}\\))'.format(re.escape(album_name)),
            '(?i)(?P<junk>\\(?{0}\\)?)'.format(re.escape(artist_name))
        ]
        item.title = remove_junk(youtube_title, artist_album_junk, YOUTUBE_TITLE_JUNK)


def remove_junk(title: str, *junk_patterns):
    new_title = title

    for pattern_list in junk_patterns:
        for pattern in pattern_list:
            match_obj = None
            if isinstance(pattern, re.Pattern):
                match_obj = pattern.search(new_title)
            elif isinstance(pattern, str):
                match_obj = re.search(pattern, title)
            if match_obj is not None:
                new_title = new_title.replace(match_obj.group('junk'), '')

    return extra_strip(new_title)


def extra_strip(string: str):
    stripped_string = string
    for pattern in EXTRA_STRIP_PATTERNS:
        match_obj = pattern.match(stripped_string)
        if match_obj is None:
            continue
        stripped_string = match_obj.group('title')
    return stripped_string.strip()



import logging
from abc import ABC, abstractmethod

class BaseApp(ABC):
    @staticmethod
    @abstractmethod
    def add_sub_parser_arguments(sub_parser):
        pass

    @abstractmethod
    def start_execution(self, arg_parser, **kwargs):
        pass

    def get_logger(self, name, level):
        logger = logging.getLogger(name)
        logger.setLevel(level)
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter('[{name}] {levelname}: {message}', style='{')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger



from setuptools import setup, find_packages

setup(
    name='ytbdl',
    version='0.0.4',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    scripts=[],

    entry_points={
        'console_scripts': [
            'ytbdl = ytbdl.application:main'
        ]
    },

    install_requires=[
        "beets==1.5.0",
        "requests>=2.0.0",
        "yt-dlp>=2021.11.10.1",
    ],

    package_data={
        'ytbdl': ['default_config.yaml']
    },

    python_requires='>=3.6.0',

    zip_safe=False,
    description='Download music with yt-dlp and autotag it with beets',
    author='Daniel Lovegrove',
    author_email='d.lovegrove11@gmail.com',
    license='MIT',
)



#pylint: disable=consider-using-f-string
from pathlib import Path
from subprocess import CalledProcessError
import shlex
import sys

from yt_dlp import main as yt_dlp_main

class SysExitSignal(Exception):
    ''' Signals a sys.exit() call
    '''
    def __init__(self, exit_code, *args, **kwargs):
        self.exit_code = exit_code
        super().__init__(*args, **kwargs)


def download_audio(album_dir: Path, extra_args: list, urls: list, logger):
    ''' Downloads one or more songs using yt-dlp into the album_dir. If the
    album_dir does not exist, yt-dlp will create it.

    To trigger yt-dlp, its main() function is called. Because the main()
    function may exit with sys.exit(), it is necessary to patch the sys.exit
    function temporarily while yt-dlp does its thing, so as not to exit from
    the ytbdl application.

    Args:
        album_dir (Path): The directory to download files into
        extra_args (list): A list of arguments to pass to yt-dlp
        urls (list): A list of URLs to download music from.
        logger: A logging object
    '''
    if extra_args:
        logger.info(
            msg='Using extra arguments for yt-dlp: {0}'.format(
            ' '.join(extra_args)
        ))
    else:
        logger.debug('No extra arguments for yt-dlp found')

    # Arguments used instead of sys.argv
    override_argv = [
        '--extract-audio',
        '--output',
        str(album_dir / '%(title)s.%(ext)s'),
        *extra_args,
        '--',
        *urls
    ]

    logger.debug(msg=(
        'Using the following yt-dlp args in place of sys.argv: {0}'.format(
        ' '.join(override_argv)
    )))

    # Patch sys.exit() so yt-dlp can't hijack the current process and exit too
    # early
    unpatched_exit = getattr(sys, 'exit')
    try:
        def patched_exit(*args, **_):
            exit_code = args[0]
            raise SysExitSignal(exit_code, 'yt-dlp exited with code {0}'.format(
                exit_code
            ))
        setattr(sys, 'exit', patched_exit)

        # Run yt-dlp's main() function
        yt_dlp_main(argv=override_argv)

    except SysExitSignal as exc:
        if exc.exit_code != 0:
            raise CalledProcessError(
                exc.exit_code, ['yt-dlp', *override_argv]
            ) from exc

    finally:
        sys.exit = unpatched_exit


def ytdl_options(value: str) -> list:
    ''' Convert a string into a set of command line arguments for yt-dlp

    Args:
        value (str): Input string received

    Returns:
        (list): A valid list of command line arguments for yt-dlp
    '''
    if not value:
        return []

    args = shlex.split(value)
    for arg in ('-x', '--extract-audio'):
        if arg in args:
            raise ValueError(
                f'The {arg} yt-dlp option is already specified for you, you do '
                'not need to add it'
            )
    for arg in ('-o', '--output'):
        if arg in args:
            raise ValueError(
                f'The {arg} yt-dlp option is already in use, you may not '
                'specify a custom output format'
            )
    return args



import shutil
from subprocess import call, CalledProcessError

from pathlib import Path
from textwrap import dedent
from argparse import RawTextHelpFormatter

from ytbdl import config, get_loaded_config_sources, config_exists, get_main_config_path
from ytbdl.apps.base import BaseApp

class ConfigApp(BaseApp):
    @staticmethod
    def add_sub_parser_arguments(sub_parser):
        config_parser = sub_parser.add_parser(name='config',
                                              formatter_class=RawTextHelpFormatter)
        config_parser.add_argument('action',
                                   choices=['path', 'edit', 'create', 'dump'],
                                   type=str,
                                   help=dedent('''\
                                       path: Show the path to the configuration file
                                       edit: Edit the configuration file. Requires editor to be set
                                       create: Create a new, empty configuration file
                                       dump: Show the contents of the configuration file
                                       '''))

    def __init__(self):
        self.logger = self.get_logger('config', 'INFO')

    def start_execution(self, arg_parser, **kwargs):
        config_action = kwargs.get('action')
        try:
            if config_action == 'create':
                if not config_exists():
                    self.create_new_config()
                else:
                    self.logger.info('Configuration file already exists here:')
                    self.print_config_sources()

            elif config_action == 'edit':
                if not config_exists():
                    self.logger.info('Create a config before continuing with:')
                    self.logger.info('ytbdl config create')
                elif 'editor' in config:
                    editor = config['editor'].get(str)
                    config_file = Path(get_loaded_config_sources()[0])
                    config_file.resolve()
                    call([editor, str(config_file)])
                else:
                    self.logger.info(
                        'To edit your configuration like this, you must first specify which '
                        'editor to use for editing. You need to add a line in your config file '
                        'similar to the following text, substituting "notepad" for the editor '
                        'you want to use:')
                    self.logger.info('\neditor: notepad')
                    self.logger.info('\nYou can find your config file here:')
                    self.print_config_sources()

            elif config_action == 'path':
                if not config_exists():
                    self.logger.info('Create a config before continuing with:')
                    self.logger.info('ytbdl config create')
                else:
                    self.print_config_sources()

            elif config_action == 'dump':
                if not config_exists():
                    self.logger.info('Create a config before continuing with:')
                    self.logger.info('ytbdl config create')
                else:
                    with open(get_loaded_config_sources()[0], 'r') as config_file:
                        for line in config_file.readlines():
                            print(line.replace('\n', ''))
            else:
                arg_parser.error(f'Invalid config action: "{config_action}"')

        except CalledProcessError as exc:
            self.logger.error(str(exc))
            self.logger.info('Try adding the selected editor to your PATH before continuing, or '
                             'just provide the full path to the editor\'s executable')
        except Exception as exc:
            self.logger.error(str(exc))

    def print_config_sources(self):
        for source in get_loaded_config_sources():
            print(source.replace('\\\\', '/'))

    def create_new_config(self):
        new_config_file = get_main_config_path()
        default_config = Path(__file__).parent.parent / 'default_config.yaml'

        if not default_config.exists():
            self.logger.info('The default_config.yaml file does not exist.')
        else:
            shutil.copy(default_config, new_config_file)
            self.logger.info(msg='Created a new configuration file at {0}.'.format(new_config_file))
